#include <stdlib.h>
#include <stdint.h>

#include <gint/display.h>
#include <gint/keyboard.h>
#include <gint/gray.h>

#include <libprof.h>

#define S3L_RESOLUTION_X 128 
#define S3L_RESOLUTION_Y 64

#define S3L_FLAT 1

#define S3L_PIXEL_FUNCTION pixfunc

#define S3L_Z_BUFFER 2

//#define S3L_SORT 2
//#define S3L_STENCIL_BUFFER 1

#define S3L_MAX_TRIANGES_DRAWN 512
#define S3L_NEAR_CROSS_STRATEGY 2

#include "../s3d/small3dlib.h"

#include "test_obj.h"

uint32_t *light, *dark;

inline void __attribute__((always_inline))
  cust_dpixel(int x, int y, color_t color){

  int offset = (y << 2) + (x >> 5);
  uint32_t mask = 1 << (~x & 31);

  if(color > 3)
    return;
  light[offset] = color & 1 ? light[offset] | mask : light[offset] & ~mask;
  dark[offset] = color & 2 ? dark[offset] | mask : dark[offset] & ~mask; 
}

void pixfunc(S3L_PixelInfo *pixel){
  int color;
  switch(pixel->triangleIndex&3){
    case 0:
      // white/dark dither
      color = ((pixel->x+pixel->y)&1)*C_DARK; 
      break;
    case 1:
      // white/black dither
      color = ((pixel->x+pixel->y)&1)*C_BLACK;
      break;
    case 2:
      // dark/black dither
      color = C_DARK + ((pixel->x+pixel->y)&1);
      break;
    case 3:
      color = C_BLACK;
      break;
  }
  cust_dpixel(pixel->x,pixel->y,color);
}

uint8_t outmap[16*16] = {
  0,0,0,0,1,2,1,0,2,0,0,0,0,0,0,0,
  0,1,1,0,2,3,1,1,0,0,1,0,1,0,0,0,
  0,2,1,0,2,4,2,2,0,0,1,2,1,0,0,0,
  0,1,1,1,1,2,3,2,1,2,1,2,0,0,0,0,
  0,1,0,0,1,0,2,3,2,3,2,1,1,0,1,0,
  0,0,1,1,1,1,3,3,4,2,3,1,2,1,1,0,
  0,1,1,0,1,1,2,3,5,3,2,2,1,0,1,1,
  1,1,0,0,0,1,2,4,6,3,2,1,2,3,1,1,
  2,1,1,1,1,2,3,4,5,4,3,2,1,1,1,0,
  1,2,1,1,1,3,4,4,4,3,2,2,1,1,0,0,
  0,1,0,0,2,2,3,2,3,2,2,1,2,0,0,0,
  0,0,0,1,1,1,2,2,3,2,2,1,1,0,0,0,
  0,0,0,0,0,0,1,1,2,1,1,1,0,0,0,0,
  0,1,1,1,0,1,1,1,1,1,0,0,0,0,0,1,
  1,1,1,0,0,1,0,1,1,0,0,0,0,0,1,2,
  0,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1
};

void heightmap_mkmodel(uint8_t *map, int size, S3L_Model3D *model){
  S3L_Unit *vertices = malloc(size*size*3*sizeof(S3L_Unit));
  S3L_Index *indices = malloc(size*size*2*sizeof(S3L_Index));
  if(!vertices || !indices)
    return;
  for(int y = 0; y < size; y++){
    for(int x = 0; x < size; x++){
      vertices[y*size*3+x*3] = x*512 - size*256;
      vertices[y*size*3+x*3+2] = y*512 - size*256;
      vertices[y*size*3+x*3+1] = map[y*size+x]*256;
    }
  }
  int i = 0;
  for(int y = 0; y < size-1; y++){
    for(int x = 0; x < size-1; x++){
      indices[i++] = y*size+x;
      indices[i++] = (y+1)*size+x+1;
      indices[i++] = y*size+x+1;
      indices[i++] = y*size+x;
      indices[i++] = (y+1)*size+x;
      indices[i++] = (y+1)*size+x+1;
    }
  }
  dclear(0);
  dprint(0,0,3,"%d",i);
  dupdate();
  getkey(); 
  S3L_model3DInit(vertices,size*size*3,indices,i,model);
  model->config.backfaceCulling = 1;
}

int main(){

  S3L_Scene scene;
 
  S3L_Model3D map_model;
  heightmap_mkmodel(outmap,16,&map_model);

  //modelModelInit();

#if S3L_Z_BUFFER 
  #if S3L_Z_BUFFER == 1
  S3L_zBuffer = malloc(sizeof(S3L_Unit)*S3L_MAX_PIXELS);
  #else
  // Moving this to a malloc()ed buffer causes a crash
  // (This buffer is only 128*64 u8's)
  S3L_zBuffer = (void*)0xe500e000 + 32;
  //S3L_zBuffer = malloc(sizeof(uint8_t)*S3L_MAX_PIXELS);
  #endif
#else
  S3L_sortArray = malloc(S3L_MAX_TRIANGES_DRAWN*sizeof(_S3L_TriangleToSort));
#endif

  S3L_sceneInit(&map_model,1,&scene);
  scene.camera.transform.translation.z = -10 * S3L_F;
  scene.camera.transform.translation.y = S3L_F / 2;

  int run = !dgray(DGRAY_ON);

  prof_init();

  int frame_time = 0;

  while(run){
    prof_t frame_timer = prof_make();
    prof_enter(frame_timer);

    dclear(C_WHITE);
    S3L_newFrame();
    dgray_getvram(&light,&dark);

    pollevent();

    if(keydown(KEY_EXIT))
      run = 0;
    if(keydown(KEY_VARS)){
      if(keydown(KEY_LEFT))
        map_model.transform.rotation.y += 10;
      if(keydown(KEY_RIGHT))
        map_model.transform.rotation.y -= 10;
      if(keydown(KEY_UP))
        map_model.transform.rotation.x += 10;
      if(keydown(KEY_DOWN))
        map_model.transform.rotation.x -= 10;
    }
    else {
      if(keydown(KEY_LEFT))
        scene.camera.transform.rotation.y += 10;
      if(keydown(KEY_RIGHT))
        scene.camera.transform.rotation.y -= 10;
      if(keydown(KEY_UP))
        scene.camera.transform.rotation.x += 10;
      if(keydown(KEY_DOWN))
        scene.camera.transform.rotation.x -= 10;
    }
    if(keydown(KEY_OPTN))
      scene.camera.transform.translation.y += 10;
    if(keydown(KEY_X2))
      scene.camera.transform.translation.y -= 10;
    if(keydown(KEY_SHIFT)){
      S3L_Vec4 trans;
      S3L_rotationToDirections(
          scene.camera.transform.rotation,
          10,
          &trans,
          NULL,
          NULL
      );
      S3L_vec3Add(&scene.camera.transform.translation,trans);
    }
    if(keydown(KEY_ALPHA)){
      S3L_Vec4 trans;
      S3L_rotationToDirections(
          scene.camera.transform.rotation,
          -10,
          &trans,
          NULL,
          NULL
      );
      S3L_vec3Add(&scene.camera.transform.translation,trans);
    }

    S3L_drawScene(scene);

    dprint(0,0,C_BLACK,"%d us", frame_time);

    dupdate();

    prof_leave(frame_timer);
    frame_time = prof_time(frame_timer);
  }

  prof_quit();

	return 1;
}
